unit GeneratorPascal;

interface

uses
  Classes, SysUtils, strutils, Source, Generator;

type

  { TGeneratorPascal }

  TGeneratorPascal = class(TGenerator)
  private
    procedure GenerateProgram(Block: TBlock;  Prog:TProgram);
    procedure GenerateFunction(ParentBlock: TBlock; FunctionDef: TFunction);
    procedure GenerateFunctionParams(ParentBlock: TBlock; Params: TFunctionParameters);
    procedure GenerateProcedure(ParentBlock: TBlock; ProcedureDef: TProcedure);
    procedure GenerateBlock(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBlockVar(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBlockConst(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBlockFunctions(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBlockProcedures(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBeginEnd(Block: TBlock; BeginEnd: TBeginEnd);
    procedure GenerateCommand(Block: TBlock; Command: TCommand);
    procedure GenerateIfThenElse(Block: TBlock; IfThenElse: TIfThenElse);
    procedure GenerateWhileDo(Block: TBlock; WhileDo: TWhileDo);
    procedure GenerateForToDo(Block: TBlock; ForToDo: TForToDo);
    procedure GenerateRepeatUntil(Block: TBlock; RepeatUntil: TRepeatUntil);
    procedure GenerateFunctionCall(Block: TBlock; FunctionCall: TFunctionCall);
    procedure GenerateProcedureCall(Block: TBlock; ProcedureCall: TProcedureCall);
    procedure GenerateAssignment(Block: TBlock; Assignment: TAssignment);
    procedure GenerateExpression(Block: TBlock; Expression: TExpression);
    procedure GenerateExpressionOperation(Block: TBlock; Expression: TExpressionOperation);
    procedure GenerateExpressionOperand(Block: TBlock; Expression: TExpressionOperand);
    procedure GenerateExpressionBrackets(Block: TBlock; Expression: TExpressionBrackets);
    procedure GenerateBreak(Block: TBlock; BreakCmd: TBreak);
    procedure GenerateContinue(Block: TBlock; ContinueCmd: TContinue);
    procedure GenerateValue(Value: TValue);
  public
    procedure Generate; override;
    constructor Create; override;
  end;


implementation

{ TGeneratorPascal }

procedure TGeneratorPascal.GenerateCommand(Block: TBlock; Command: TCommand);
begin
  if Command is TBeginEnd then GenerateBeginEnd(Block, TBeginEnd(Command))
  else if Command is TFunctionCall then GenerateFunctionCall(Block, TFunctionCall(Command))
  else if Command is TProcedureCall then GenerateProcedureCall(Block, TProcedureCall(Command))
  else if Command is TAssignment then GenerateAssignment(Block, TAssignment(Command))
  else if Command is TIfThenElse then GenerateIfThenElse(Block, TIfThenElse(Command))
  else if Command is TWhileDo then GenerateWhileDo(Block, TWhileDo(Command))
  else if Command is TRepeatUntil then GenerateRepeatUntil(Block, TRepeatUntil(Command))
  else if Command is TForToDo then GenerateForToDo(Block, TForToDo(Command))
  else if Command is TBreak then GenerateBreak(Block, TBreak(Command))
  else if Command is TContinue then GenerateContinue(Block, TContinue(Command))
  else if Command is TEmptyCommand then
  else raise Exception.Create('Unsupported command type');
end;

procedure TGeneratorPascal.GenerateIfThenElse(Block: TBlock; IfThenElse: TIfThenElse);
begin
  AddText('if ');
  GenerateExpression(Block, IfThenElse.Expression);
  AddText(' then ');
  GenerateCommand(Block, IfThenElse.CommandThen);
  if Assigned(IfThenElse.CommandElse) and not (IfThenElse.CommandElse is TEmptyCommand) then begin
    AddText(' else ');
    GenerateCommand(Block, IfThenElse.CommandElse);
  end;
end;

procedure TGeneratorPascal.GenerateWhileDo(Block: TBlock; WhileDo: TWhileDo);
begin
  AddText('while ');
  GenerateExpression(Block, WhileDo.Expression);
  AddText(' do ');
  GenerateCommand(Block, WhileDo.Command);
end;

procedure TGeneratorPascal.GenerateForToDo(Block: TBlock; ForToDo: TForToDo);
begin
  AddText('for ');
  AddText(ForToDo.VariableRef.Name);
  AddText(' := ');
  GenerateExpression(Block, ForToDo.ExpressionFrom);
  AddText(' to ');
  GenerateExpression(Block, ForToDo.ExpressionTo);
  AddText(' do ');
  GenerateCommand(Block, ForToDo.Command);
end;

procedure TGeneratorPascal.GenerateRepeatUntil(Block: TBlock;
  RepeatUntil: TRepeatUntil);
var
  I: Integer;
begin
  AddTextLine('repeat');
  Indent := Indent + 1;
  for I := 0 to RepeatUntil.Commands.Count - 1 do begin
    GenerateCommand(Block, RepeatUntil.Commands[I]);
    AddTextLine(';');
  end;
  Indent := Indent - 1;
  AddText('until ');
  GenerateExpression(Block, RepeatUntil.Expression);
end;

procedure TGeneratorPascal.GenerateFunctionCall(Block: TBlock;
  FunctionCall: TFunctionCall);
var
  I: Integer;
begin
  AddText(FunctionCall.FunctionDef.Name);
  if FunctionCall.Params.Count > 0 then begin
    AddText('(');
    for I := 0 to FunctionCall.Params.Count - 1 do
      GenerateExpression(Block, FunctionCall.Params[I]);
    AddText(')');
  end;
end;

procedure TGeneratorPascal.GenerateProcedureCall(Block: TBlock;
  ProcedureCall: TProcedureCall);
var
  I: Integer;
begin
  AddText(ProcedureCall.ProcedureDef.Name);
  if ProcedureCall.Params.Count > 0 then begin
    AddText('(');
    for I := 0 to ProcedureCall.Params.Count - 1 do
      GenerateExpression(Block, ProcedureCall.Params[I]);
    AddText(')');
  end;
end;

procedure TGeneratorPascal.GenerateAssignment(Block: TBlock; Assignment: TAssignment);
begin
  AddText(Assignment.Variable.Name);
  AddText(' := ');
  GenerateExpression(Block, Assignment.Expression);
end;

procedure TGeneratorPascal.GenerateExpression(Block: TBlock; Expression: TExpression);
begin
  if Expression is TExpressionOperation then
    GenerateExpressionOperation(Block, TExpressionOperation(Expression))
  else
  if Expression is TExpressionOperand then
    GenerateExpressionOperand(Block, TExpressionOperand(Expression))
  else
  if Expression is TExpressionBrackets then
    GenerateExpressionBrackets(Block, TExpressionBrackets(Expression))
  else raise Exception.Create('Unknown expression class.');
end;

procedure TGeneratorPascal.GenerateExpressionOperation(Block: TBlock;
  Expression: TExpressionOperation);
var
  I: Integer;
begin
  for I := 0 to Expression.Items.Count - 1 do begin
    if I > 0 then begin
      AddText(' ');
      AddText(ExpressionOperatorText[Expression.Operation]);
      AddText(' ');
    end;
    GenerateExpression(Block, Expression.Items[I]);
  end;
end;

procedure TGeneratorPascal.GenerateExpressionOperand(Block: TBlock;
  Expression: TExpressionOperand);
begin
  case Expression.OperandType of
    otFunctionCall: GenerateFunctionCall(Block, Expression.FunctionCall);
    otConstantDirect: GenerateValue(Expression.ConstantDirect.Value);
    otConstantRef: AddText(Expression.ConstantRef.Name);
    otVariableRef: AddText(Expression.VariableRef.Name);
    else raise Exception.Create('Unsupported exception operand type.');
  end;
end;

procedure TGeneratorPascal.GenerateExpressionBrackets(Block: TBlock;
  Expression: TExpressionBrackets);
begin
  AddText('(');
  GenerateExpression(Block, Expression.Expression);
  AddText(')');
end;

procedure TGeneratorPascal.GenerateBreak(Block: TBlock; BreakCmd: TBreak);
begin
  AddText('break');
end;

procedure TGeneratorPascal.GenerateContinue(Block: TBlock;
  ContinueCmd: TContinue);
begin
  AddText('continue');
end;

procedure TGeneratorPascal.GenerateValue(Value: TValue);
begin
  if Value is TValueBoolean then begin
    if TValueBoolean(Value).Value then AddText('True') else AddText('False');
  end else if Value is TValueString then AddText('''' + StringReplace(TValueString(Value).Value, '''', '''''', [rfReplaceAll]) + '''')
  else if Value is TValueInteger then AddText(IntToStr(TValueInteger(Value).Value))
  else raise Exception.Create('Unsupported value type.');
end;

procedure TGeneratorPascal.GenerateProgram(Block: TBlock; Prog: TProgram);
begin
  if Prog.Name <> '' then AddTextLine('program ' + Prog.Name + ';');
  AddTextLine('{$mode delphi}');
  AddTextLine('');
  AddTextLine('uses');
  AddTextLine('  SysUtils;');
  AddTextLine('');
  GenerateBlock(Block, Prog.Block);
  AddTextLine('.');
end;

procedure TGeneratorPascal.GenerateFunction(ParentBlock: TBlock;
  FunctionDef: TFunction);
begin
  AddText('function ' + FunctionDef.Name);
  GenerateFunctionParams(ParentBlock, FunctionDef.Params);
  if Assigned(FunctionDef.ResultType) then begin
    AddText(': ');
    AddText(FunctionDef.ResultType.Name);
  end;
  AddTextLine(';');
  if FunctionDef.InternalName <> '' then begin
    AddTextLine('begin');
    Indent := Indent + 1;
    if FunctionDef.InternalName = 'IntToStr' then AddTextLine('return SysUtils.IntToStr(Value);')
    else if FunctionDef.InternalName = 'StrToInt' then AddTextLine('return SysUtils.StrToInt(Value);')
    else if FunctionDef.InternalName = 'BoolToStr' then AddTextLine('return SysUtils.BoolToStr(Value);')
    else if FunctionDef.InternalName = 'StrToBool' then AddTextLine('return SysUtils.StrToBool(Value);');
    Indent := Indent - 1;
    AddTextLine('end;');
  end else begin
    GenerateBlock(ParentBlock, FunctionDef.Block);
    AddTextLine(';');
  end;
end;

procedure TGeneratorPascal.GenerateFunctionParams(ParentBlock: TBlock;
  Params: TFunctionParameters);
var
  I: Integer;
begin
  if Params.Count > 0 then begin
    AddText('(');
    for I := 0 to Params.Count - 1 do begin
      if Params[I].Kind = pkVar then AddText('var ');
      if Params[I].Kind = pkConst then AddText('const ');
      AddText(Params[I].Name);
      AddText(': ');
      AddText(Params[I].TypeRef.Name);
      if I > 0 then AddText(', ');
    end;
    AddText(')');
  end;
end;

procedure TGeneratorPascal.GenerateProcedure(ParentBlock: TBlock;
  ProcedureDef: TProcedure);
begin
  AddText('procedure ' + ProcedureDef.Name);
  GenerateFunctionParams(ParentBlock, ProcedureDef.Params);
  AddTextLine(';');
  if ProcedureDef.InternalName <> '' then begin
    AddTextLine('begin');
    Indent := Indent + 1;
    if ProcedureDef.InternalName = 'WriteLn' then AddTextLine('System.WriteLn(Text);')
    else if ProcedureDef.InternalName = 'Write' then AddTextLine('System.Write(Text);')
    else if ProcedureDef.InternalName = 'ReadLn' then AddTextLine('System.ReadLn(Text);')
    else if ProcedureDef.InternalName = 'Read' then AddTextLine('System.Read(Text);');
    Indent := Indent - 1;
    AddTextLine('end;');
  end else begin
    GenerateBlock(ParentBlock, ProcedureDef.Block);
    AddTextLine(';');
  end;
end;

procedure TGeneratorPascal.GenerateBlock(ParentBlock: TBlock; Block: TBlock);
begin
  GenerateBlockConst(ParentBlock, Block);
  GenerateBlockVar(ParentBlock, Block);
  GenerateBlockProcedures(ParentBlock, Block);
  GenerateBlockFunctions(ParentBlock, Block);
  GenerateBeginEnd(ParentBlock, Block.BeginEnd);
end;

procedure TGeneratorPascal.GenerateBlockVar(ParentBlock: TBlock; Block: TBlock);
var
  I: Integer;
  Variable: TVariable;
  VarCount: Integer;
begin
  VarCount := 0;
  for I := 0 to Block.Variables.Count - 1 do
    if not Block.Variables[I].Internal then Inc(VarCount);

  if VarCount > 0 then begin
    AddTextLine('var');
    Indent := Indent + 1;
    for I := 0 to Block.Variables.Count - 1 do
    if not Block.Variables[I].Internal then begin
      Variable := Block.Variables[I];
      AddTextLine(Variable.Name + ': ' + Variable.TypeRef.Name + ';');
    end;
    Indent := Indent - 1;
    AddTextLine;
  end;
end;

procedure TGeneratorPascal.GenerateBlockConst(ParentBlock: TBlock; Block: TBlock);
var
  I: Integer;
  Constant: TConstant;
begin
  if Block.Constants.Count > 0 then begin
    AddTextLine('const');
    Indent := Indent + 1;
    for I := 0 to Block.Constants.Count - 1 do begin
      Constant := Block.Constants[I];
      AddText(Constant.Name + ': ' + Constant.TypeRef.Name + ' = ');
      GenerateValue(Constant.Value);
      AddTextLine(';');
    end;
    Indent := Indent - 1;
    AddTextLine;
  end;
end;

procedure TGeneratorPascal.GenerateBeginEnd(Block: TBlock; BeginEnd: TBeginEnd);
var
  I: Integer;
begin
  AddTextLine('begin');
  Indent := Indent + 1;
  for I := 0 to BeginEnd.Commands.Count - 1 do begin
    GenerateCommand(Block, BeginEnd.Commands[I]);
    AddTextLine(';');
  end;
  Indent := Indent - 1;
  AddText('end');
end;

procedure TGeneratorPascal.GenerateBlockFunctions(ParentBlock: TBlock;
  Block: TBlock);
var
  I: Integer;
begin
  for I := 0 to Block.Functions.Count - 1 do begin
    GenerateFunction(ParentBlock, Block.Functions[I]);
    AddTextLine;
  end;
end;

procedure TGeneratorPascal.GenerateBlockProcedures(ParentBlock: TBlock;
  Block: TBlock);
var
  I: Integer;
begin
  for I := 0 to Block.Procedures.Count - 1 do begin
    GenerateProcedure(ParentBlock, Block.Procedures[I]);
    AddTextLine;
  end;
end;

procedure TGeneratorPascal.Generate;
begin
  Output := '';
  GenerateProgram(Prog.SystemBlock, Prog);
end;

constructor TGeneratorPascal.Create;
begin
  inherited;
  Name := 'Pascal';
  FileExt := '.pas';
end;

end.

